/*
 * Copyright 2002, Tobias Huerlimann. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <types.h>
#include <string.h>

#define INCL_DOSERRORS
#define INCL_DOSNMPIPES
#define INCL_DOSPROCESS
#define INCL_DOSFILEMGR
#include <os2.h>

#include <rndos2.h>

#define BUFSIZE			256

#define PIPENAME_RND		"\\PIPE\\RANDOM"
#define PIPENAME_SRND		"\\PIPE\\SRANDOM"
#define PIPENAME_URND		"\\PIPE\\URANDOM"
#define PIPENAME_PRND		"\\PIPE\\PRANDOM"
#define PIPENAME_ARND		"\\PIPE\\ARANDOM"

int			verbose = 0;
char*			progname;
TID			thread_ids[RND_NODEV];
char*			pipenames[RND_NODEV] =
				{ PIPENAME_RND, PIPENAME_SRND, PIPENAME_URND,
				  PIPENAME_PRND, PIPENAME_ARND };

void usage(void);
void APIENTRY pipe_thrd(ULONG);

int main(int argc, char** argv) {
	APIRET		rc = NO_ERROR;
	int		i, j;
	int		arglen;

	progname = argv[0];

	/* Some simple command line parsing */
	for (i = 1; i < argc; i++) {
		if (*argv[i] != '-') {
			usage();
			return 0;
		}
		arglen = strlen(argv[i]);
		for (j = 1; j < arglen; j++)
			switch (*(argv[i] + j)) {
				case 'v':
					verbose++;
					break;
				case 'h':
				default:
					usage();
					return 0;
			}
	}

	/* Create the thread that provide the pipes */
	for (i = RND_RND; i < RND_NODEV; i++) {
		if (verbose >= 2)
			fprintf(stderr, "%s: Creating pipe thread for random device %u ...\n", progname, i);
		rc = DosCreateThread(&thread_ids[i], &pipe_thrd, i, CREATE_READY | STACK_SPARSE, 16384);
		if (rc != NO_ERROR) {
			fprintf(stderr, "%s: DosCreateThread() failed (rc = %u)\n", progname, rc);
			return rc;
		}
	}

	if (verbose >= 1)
		fprintf(stderr, "%s: Initialization finished, main() begins infinite loop\n", progname);
	while (TRUE)
		DosSleep(10000);

	return 0;
}

void usage() {
	printf("%s - rndos2 daemon\n\n", progname);
	printf("Usage: %s [-vh]\n", progname);
	printf("\t-v\tIncrease verbosity level\n");
	printf("\t-h\tShow this help and exit\n");
}

void pipe_thrd(ULONG rnd_dev) {
	APIRET		rc = NO_ERROR;
	HPIPE		pipe;
	ULONG		written;
	char		buf[BUFSIZE];

	if (verbose >= 2)
		fprintf(stderr, "%s: Creating named pipe %s for random device %u ...\n",
			progname, pipenames[rnd_dev], rnd_dev);
	rc = DosCreateNPipe(pipenames[rnd_dev], &pipe,
		NP_ACCESS_OUTBOUND | NP_NOINHERIT | NP_NOWRITEBEHIND,
		NP_WAIT | NP_TYPE_BYTE | NP_READMODE_BYTE | NP_UNLIMITED_INSTANCES,
		BUFSIZE, 0UL, 0UL);
	if (rc != NO_ERROR) {
		fprintf(stderr, "%s: DosCreateNPipe() failed (rc = %u)\n", progname, rc);
		fprintf(stderr, "%s: Warning: %s not available\n", progname, pipenames[rnd_dev]);
		DosExit(EXIT_THREAD, rc);
	}

	if (verbose >= 3)
		fprintf(stderr, "%s: Opening random device %u ...\n",
			progname, rnd_dev);
	if (rndos2_open(rnd_dev) != 0) {
		rc = 1;
		goto thread_cleanup;
	}

	while (TRUE) {
		if (verbose >= 3)
			fprintf(stderr, "%s: Connecting pipe for random device %u (\"%s\")...\n",
				progname, rnd_dev, pipenames[rnd_dev]);
		rc = DosConnectNPipe(pipe);
		if (rc != NO_ERROR) {
			fprintf(stderr, "%s: DosConnectNPipe() failed (rc = %u)\n", progname, rc);
			goto thread_cleanup;
		}
		do {
			if (verbose >= 2)
				fprintf(stderr, "%s: Reading %u Byte for output on %s ...\n",
				progname, BUFSIZE, pipenames[rnd_dev]);
			if (rndos2_read(rnd_dev, (char*)&buf, sizeof(buf)) != 0) {
				fprintf(stderr, "%s: rndos2_read(%u, ...) failed\n", progname, rnd_dev);
				rc = 1;
				goto thread_cleanup;
			}
			rc = DosWrite(pipe, &buf, sizeof(buf), &written);
		} while (rc == NO_ERROR);
		if ((rc != ERROR_BROKEN_PIPE) && (rc != ERROR_DISCARDED)) {
			fprintf(stderr, "%s: DosWrite() failed (rc = %u)\n", progname, rc);
			goto thread_cleanup;
		}
		rc = DosDisConnectNPipe(pipe);
		if (rc != NO_ERROR) {
			fprintf(stderr, "%s: DosDisConnectNPipe() failed (rc = %u)\n", progname, rc);
			goto thread_cleanup;
		}
	}

	rc = NO_ERROR;
thread_cleanup:
	if (verbose >= 3)
		fprintf(stderr, "%s: Closing pipe %s ...\n", progname, pipenames[rnd_dev]);
	/* No rc checking here, we do not care anymore */
	DosDisConnectNPipe(pipe);
	rndos2_close(rnd_dev);
	DosClose(pipe);
	DosExit(EXIT_THREAD, rc);
}

